
/***********************************************************************

doskillthread - DosKillThread stress test

Copyright (c) 2017 Steven Levine and Associates, Inc.
All rights reserved.

2017-12-15 SHL Baseline
2017-12-20 SHL Update
2017-12-25 SHL Support DosCreateThread fail

Usage		See Usage()

Exit status	0	OK
		1-253	OS error
		254	Fatal error
		255	Usage error

Build with
  WatAppDev
  wcl386 doskillthread slawat.lib
    see doskillthread.mak

***********************************************************************/

#include <stdlib.h>			// NULL _doserrno
#include <stdio.h>
#include <stdarg.h>			// va_list
#include <fcntl.h>
#include <string.h>
#include <ctype.h>			// isdidit

#include <unistd.h>			// getopt

// #include <shldef.h>

#define INCL_DOS
#define INCL_DOSPROCESS			// DosGetInfoBlocks
#define INCL_DOSERRORS
#include <os2.h>

#include <shlstdli.h>			// Errno2Msg

#define	PGM_ID		"doskillthread"
#define VERSION		1
#define	REVISION	0

static char * Copyright[] = {
  "Copyright (c) 2017 Steven Levine and Associates, Inc.",
  "All rights reserved."
};

struct
{
  UINT cDebug;
  UINT cVerbose;
  UINT cWorkerThreads;			// To create
  UINT cKillDelay;			// Between kills, mSec
  UINT cKillLimit;			// Before shutting down
} Opts = {
  0,					// cDebug
  0,					// cVerbose
  5,					// cWorkerThreads
  2000,					// cKillDelay
  10					// cKillLimit
};

static CHAR szCmdName[CCHMAXPATH];

typedef struct
{
  UINT	tid;
} ThreadInfo;

void main(int argc, char **argv);

VOID CreateThreads();
VOID Init(INT argc, CHAR **argv);
ULONG _System ExceptionHandler(PEXCEPTIONREPORTRECORD       p1,
			       PEXCEPTIONREGISTRATIONRECORD p2,
			       PCONTEXTRECORD               p3,
			       PVOID                        pv );
VOID KillThreads();

VOID Help(VOID);
VOID Usage(CHAR *pszFmt,...);
FNTHREAD WorkerThread;

VOID Dos_Error(ULONG apiret,
	       PCSZ pszSrcFile, UINT cSrcLineNo, PCSZ pszFmt, ...);
VOID Fatal(CHAR *pszFmt,...);
VOID FatalIO(CHAR *pszAction, CHAR *pszPathName);
VOID Runtime_Error(PCSZ pszSrcFile, UINT cSrcLineNo, PCSZ pszFmt, ...);
VOID Shut(SHORT sStatus);
VOID WriteMsg(PCSZ pszFmt, ...);

#if 1 // Used by WriteMsg - not really needed because we are single thread

// SMP safe inc/dec/cmpxchg
extern void SMPSafeInc(volatile ULONG* pul);
#pragma aux SMPSafeInc = \
	"lock inc near ptr 0[eax]" \
	parm [eax] modify exact []

extern void SMPSafeDec(volatile ULONG* pul);
#pragma aux SMPSafeDec = \
	"lock dec near ptr 0[eax]" \
	parm [eax] modify exact []

// Return true if xchg ok
extern int SMPCmpXchg(volatile ULONG * pul, ULONG old, ULONG new);
#pragma aux SMPCmpXchg = \
	"lock cmpxchg near ptr 0[ebx], edx" \
	"setz al" \
	"movzx eax, al" \
	parm [ebx] [eax] [edx] \
	modify exact [eax] \
	value [eax]
#endif

UINT cKillRequests;
UINT cThreadsCreated;
UINT cThreadsEnded;
UINT cThreadsTerminated;
BOOL fAsyncContinue;		// XCPT_CONTINUE_EXECUTION if XCPT_ASYNC_PROCESS_TERMINATE
BOOL fShuttingdown;

/**
 * main
 * @param argc is argument count
 * @param argv is argument array
 * @returns 0 if ok otherwise error code
 */

VOID main(INT argc, CHAR **argv)
{
  Init(argc, argv);

  CreateThreads();

  puts("Sleeping 1 second for threads to settle");
  DosSleep(1000);

  KillThreads();

  puts("Quitting time");

  // And shut down
  Shut(0);

} // main

#if 0 // Template
/*
 * Do some work
 */

VOID DoWork();
VOID DoWork()
{
  puts("Working");

} // DoWork
#endif

/*
 * Create threads
 */

VOID CreateThreads()
{
  UINT i;
  ThreadInfo ti;
  APIRET apiret;
  TID tid;

  WriteMsg("Creating %u threads", Opts.cWorkerThreads);

  // We are thread 1 - worker threads are 2..N
  for (ti.tid = 2; ti.tid <= Opts.cWorkerThreads + 1; ti.tid++) {
    if (Opts.cVerbose)
      WriteMsg("Creating thread %u", ti.tid);
    apiret = DosCreateThread(&tid,
			     WorkerThread,
			     (ULONG)&ti,
			     CREATE_READY | STACK_COMMITTED,
			     32 * 1024);

    if (apiret != NO_ERROR) {
      Dos_Error(apiret, __FILE__, __LINE__, "DosCreateThread");
      if (apiret == ERROR_MAX_THRDS_REACHED)
	break;
    }
    else
      cThreadsCreated++;
     DosSleep(10);			// Let thread use ThreadInfo
  } // for
  WriteMsg("Created %u threads", cThreadsCreated);
} // CreateThreads

/**
 * Handle DosKillThread exceptions
 */

ULONG _System ExceptionHandler(PEXCEPTIONREPORTRECORD       p1,
			       PEXCEPTIONREGISTRATIONRECORD p2,
			       PCONTEXTRECORD               p3,
			       PVOID                        pv)
{
  PTIB ptib;
  PPIB ppib;
  DosGetInfoBlocks(&ptib, &ppib);
  if (Opts.cVerbose ||
      !fAsyncContinue ||
      (p1->ExceptionNum != XCPT_ASYNC_PROCESS_TERMINATE &&
       p1->ExceptionNum != XCPT_UNWIND))
  {
    WriteMsg("Thread %u exception handler entered for %x", ptib->tib_ptib2->tib2_ultid, p1->ExceptionNum);
  }
  switch ( p1->ExceptionNum) {
  case XCPT_SIGNAL:
  {
    switch( p1->ExceptionInfo[0] ) {
      case XCPT_SIGNAL_INTR:
      { WriteMsg("XCPT_SIGNAL_INTR");
	fShuttingdown = 1;
	break;
      }
      case XCPT_SIGNAL_KILLPROC:
      { WriteMsg("XCPT_SIGNAL_KILLPROC");
	fShuttingdown = 1;
	break;
      }
      case XCPT_SIGNAL_BREAK:
      {
	WriteMsg("XCPT_SIGNAL_BREAK");
	fShuttingdown = 1;
      }
      default:
	WriteMsg("SIGNAL 0x%x unknown", p1->ExceptionInfo[0]);

    } // switch
    break;
  }
  case XCPT_PROCESS_TERMINATE:
  {
    WriteMsg("Process Terminate");
    break;
  }
  case XCPT_ASYNC_PROCESS_TERMINATE:
  {
  // This can hang because WriteMsg not quite 100% SMP safe
  if (Opts.cVerbose >= 2)
      WriteMsg("XCPT_ASYNC_PROCESS_TERMINATE");
    // FIXME to report cs:eip
    if (ptib->tib_ptib2->tib2_ultid != 1 && !fShuttingdown && fAsyncContinue)
      return XCPT_CONTINUE_EXECUTION;	// Allow thread to continue
    if (fShuttingdown)
      cThreadsTerminated++;
    break;
  }
  case XCPT_UNWIND:
  {
    if (Opts.cVerbose >= 2 || !fAsyncContinue || !fShuttingdown)
      WriteMsg("XCPT_UNWIND");
    break;
  }
  } // switch
  return XCPT_CONTINUE_SEARCH;		// Exception not resolved...
}

/*
 * Kill threads randomly
 */

VOID KillThreads()
{
  EXCEPTIONREGISTRATIONRECORD RegRec  = {0};	// Exception Registration Record
  TID tid;
  APIRET apiret;
  ULONG cThreadsLeft = cThreadsCreated;
  ULONG ulTimes;

  RegRec.ExceptionHandler = (ERR)ExceptionHandler;
  apiret = DosSetExceptionHandler(&RegRec);
  if (apiret != NO_ERROR)
    Dos_Error(apiret, __FILE__, __LINE__, "DosSetExceptionHandler");

  apiret = DosSetSignalExceptionFocus(SIG_SETFOCUS, &ulTimes);
  if (apiret != NO_ERROR)
    Dos_Error(apiret, __FILE__, __LINE__, "DosSetSignalExceptionFocus");

  WriteMsg("Invoking DosKillThread %u times, sleeping %u msec between kills",
	   Opts.cKillLimit, Opts.cKillDelay);
  fAsyncContinue = 1;

  for (;;) {
    tid = (rand() % cThreadsCreated) + 2;
    if (Opts.cVerbose)
      WriteMsg("Killing tid %u", tid);
    apiret = DosKillThread(tid);
    cKillRequests++;
    // ERROR_INVALID_THREADID OK if shutting down
    if (apiret != NO_ERROR && (!fShuttingdown || apiret != ERROR_INVALID_THREADID))
      Dos_Error(apiret, __FILE__, __LINE__, "DosKillThread failed for tid %u", tid);
    if (cKillRequests >= Opts.cKillLimit) {
      ULONG ul;
#if 1
      ul = cThreadsCreated - (cThreadsEnded + cThreadsTerminated);
      if (ul < cThreadsLeft) {
	cThreadsLeft = ul;
	WriteMsg("KillThreads waiting for %u threads to die", cThreadsLeft);
      }
#endif
      fShuttingdown = 1;
    }
    if (cThreadsEnded + cThreadsTerminated >= cThreadsCreated)
      break;				//  All worker threads gone
    if (Opts.cKillDelay)
      DosSleep(Opts.cKillDelay);
  } // for

  WriteMsg("KillThreads exiting after %u kill requests", cKillRequests);
  WriteMsg("KillThreads cThreadsTerminated %u cThreadsEnded %u", cThreadsTerminated, cThreadsEnded);

  apiret = DosUnsetExceptionHandler(&RegRec);
  if (apiret != NO_ERROR)
    Dos_Error(apiret, __FILE__, __LINE__, "DosUnsetExceptionHandler");

} // KillThreads

/*
 * Worker thread
 */

VOID WorkerThread(ULONG ulArg)
{
  ThreadInfo ti = *(ThreadInfo *)ulArg;	// Make local copy
  EXCEPTIONREGISTRATIONRECORD RegRec  = {0};	// Exception Registration Record
  APIRET apiret;

  RegRec.ExceptionHandler = (ERR)ExceptionHandler;
  apiret = DosSetExceptionHandler(&RegRec);
  if (apiret != NO_ERROR)
    Dos_Error(apiret, __FILE__, __LINE__, "DosSetExceptionHandler");

  if (Opts.cVerbose)
    WriteMsg("Thread %u running", ti.tid);

  for (;;) {
    // Fails with ERROR_TS_WAKEUP if woken up by DosKillThread
    apiret = DosSleep(60 * 1000);
    if (fShuttingdown)
      break;
    if (apiret == ERROR_TS_WAKEUP) {
      if (Opts.cVerbose)
	WriteMsg("Thread %u DosSleep terminated with ERROR_TS_WAKEUP", ti.tid);
    }
    else if (apiret != NO_ERROR)
      Dos_Error(apiret, __FILE__, __LINE__, "DosSleep");
    if (cKillRequests >= Opts.cKillLimit)
      break;
  } // for

  apiret = DosUnsetExceptionHandler(&RegRec);
  if (apiret != NO_ERROR)
    Dos_Error(apiret, __FILE__, __LINE__, "DosUnsetExceptionHandler");

  WriteMsg("Thread %u exiting", ti.tid);

  cThreadsEnded++;

} // WorkerThread

/*
 * Initialize
 * @param argc is Argument count
 * @param argv is argument array
 */

VOID Init(INT argc, CHAR **argv)
{
  CHAR		*pszArg;
  USHORT	cArgLen;
  SHORT		sOptChr;
  PSZ		psz;

  if (argv[0]) {
    psz = strrchr(argv[0], '\\');
    if (!psz)
      psz = PGM_ID;
    strcpy(szCmdName, psz + 1);
    psz = strrchr(szCmdName, '.');
    if (psz)
      *psz = 0;
    strlwr(szCmdName);
  }

  // Scan switches - : prefex suppresses getopt messages

  while ((sOptChr = getopt(argc, argv, ":dhk:t:vVw:")) != -1 )	{
    switch (sOptChr) {
    case '?':				// Getopt has already complained
      if (sOptChr != optopt)
	Usage("Option %c unexpected.", optopt);
      // Drop thru
    case 'h':
      Help();
      break;
    case 'd':
      Opts.cDebug++;
      break;
    case 'k':
      if (!isdigit(*optarg))
	Usage("kill limit %s must be a non-negative integer.", optarg);
      Opts.cKillLimit = atoi(optarg);
      if (!Opts.cKillLimit)
	Usage("kill limit must be non-zero.");
      break;
    case 't':
      if (!isdigit(*optarg))
	Usage("kill delay %s must be a positive integer.", optarg);
      Opts.cKillDelay = atoi(optarg);
#if 0
      if (!Opts.cKillDelay)
	Usage("kill delay must be non-zero.");
#endif
      break;
    case 'v':
      Opts.cVerbose++;
      break;
    case 'V':
      printf("%s version %u.%u", szCmdName, VERSION, REVISION);
      Shut(0);
      break;
    case 'w':
      if (!isdigit(*optarg))
	Usage("thread count %s must be a positive integer.", optarg);
      Opts.cWorkerThreads = atoi(optarg);
      if (!Opts.cWorkerThreads)
	Usage("thread count must be non-zero.");
      break;
    default:
      Usage("Unknown switch: %c.", sOptChr);
    } // switch
  } // while switches

  // Scan arguments

  for (;optind < argc; optind++) {
    pszArg = argv[optind];
    cArgLen = strlen(pszArg);

    if (cArgLen == 0)
      Usage("Null argument.");

    // Assume file name

    Usage("Unexpected argument.");

  } // while args

} // Init

/**
 * Display usage help
 * @param pszFmt is format string, message or NULL
 */

#define HELP_MSG1 \
  "DosKillThread stress test. v%u.%u.\n" \
  "\n"

#define HELP_MSG \
  "%s [-d] [-h] [-k limit] [-t delay] [-V] [-w threads] [-?]\n"

#define HELP_MSG2 \
  "\n" \
  "  -d          Enable debug mode.\n" \
  "  -h -?       Display this message.\n" \
  "  -t msecs    Kill delay (msec), default is 2000 msec.\n" \
  "  -k limit    Kill limit, default is 5.\n" \
  "  -v          Be verbose.\n" \
  "  -V          Display version.\n" \
  "  -w threads  Worker threads to create, default is 5.\n" \
  "\n"

VOID Help(VOID)
{
  printf("\n" HELP_MSG1 HELP_MSG HELP_MSG2, VERSION, REVISION, szCmdName);
  printf("%s\n%s\n", Copyright[0], Copyright[1]);
  exit(255);

} // Help

/**
 * Report usage error to stdout
 * @param pszFmt is format string, message or NULL
 */

VOID Usage(CHAR *pszFmt, ...)
{
  va_list	pVA;
  va_start(pVA, pszFmt);		// Point at arguments

  // If message passed complain
  if (pszFmt != NULL) {
    fputc('\n', stdout);
    vfprintf(stdout, pszFmt, pVA);
    fputc('\n', stdout);
  }
  fprintf(stdout, HELP_MSG, szCmdName);
  exit(255);

} // Usage

//========================================================================

/*
 * Report fatal error
 * @param pszFmt is format string or message
 */

VOID Fatal(CHAR *pszFmt, ...)
{
  va_list	pVA;

  // Point at arguments
  va_start(pVA, pszFmt);

  // Complain
  fputs("\n\a" PGM_ID ": ", stderr);
  vfprintf(stderr, pszFmt, pVA);

  exit(254);

} // Fatal

/*
 * Report fatal IO error
 * @param pszAction is action that failed
 * @param pszPathName is path name associated with error or NULL
 */

VOID FatalIO(CHAR *pszAction, CHAR *pszPathName)
{
  CHAR	szMsg[60];

  if (pszPathName == NULL)
    Fatal("Can not %s. %s.", pszAction, Errno2Msg(szMsg));
  else
    Fatal("Can not %s %s. %s.", pszAction, pszPathName, Errno2Msg(szMsg));

} // FatalIO

/**
 * Report Dos...() error using passed message string
 */

VOID Dos_Error(ULONG apiret,
	       PCSZ pszSrcFile, UINT cSrcLineNo, PCSZ pszFmt, ...)
{
  CHAR szMsg[4096];
  ULONG Class;				// Error class
  ULONG action;				// Error action
  ULONG Locus;				// Error location
  ULONG ulMsgLen;
  CHAR *pszMsgStart;
  CHAR *psz;
  va_list va;

  if (!apiret)
    return;				// Should not have been called

  // Format caller's message
  va_start(va, pszFmt);
  szMsg[sizeof(szMsg) - 1] = 0;
  vsprintf(szMsg, pszFmt, va);
  va_end(va);

  if (szMsg[sizeof(szMsg) - 1]) {
    fprintf(stdout, "Buffer overflow in Dos_Error - need %u bytes.\n",
	    strlen(szMsg) + 1);
    fflush(stdout);
  }

  if (strchr(szMsg, ' ') == NULL)
    strcat(szMsg, " failed.");		// Assume simple function name

  DosErrClass(apiret, &Class, &action, &Locus);

  #define IDS_DOSERR1TEXT \
	"\nModule: %s Line number: %u\n" \
	"OS/2 error: %d\nClass: %d\nAction: %d\nLocation: %d"

  sprintf(szMsg + strlen(szMsg),
	  IDS_DOSERR1TEXT,
	  pszSrcFile,
	  cSrcLineNo,
	  apiret,
	  Class,
	  action,
	  Locus);
	  // GetPString(IDS_ERRCLASS1TEXT + (Class - 1)),
	  // GetPString(IDS_ERRACTION1TEXT + (action - 1)),
	  // GetPString(IDS_ERRLOCUS1TEXT + (Locus - 1));
  pszMsgStart = szMsg + strlen(szMsg) + 1;
  // Get message leaving space for NL separator
  if (!DosGetMessage
      (NULL, 0L, (PCHAR) pszMsgStart + 1, 1024, apiret, "OSO001.MSG", &ulMsgLen)
      || !DosGetMessage(NULL, 0L, (PCHAR) pszMsgStart + 1, 1024, apiret,
			"OSO001H.MSG", &ulMsgLen)) {
    // Got message
    pszMsgStart[ulMsgLen + 1] = 0;	// Terminate
    *(pszMsgStart - 1) = '\n';		// Stuff NL before message text
    *pszMsgStart = '\"';		// Prefix message text with quote

    psz = pszMsgStart + ulMsgLen;	// Point at last CHAR
    // Chop trailing NL CR TAB
    while (*psz &&
	   (*psz == '\r' || *psz == '\n' || *psz == ' ' || *psz == '\t')) {
      *psz-- = 0;
    }
    strcat(psz, "\"");			// Append trailing quote

    // Convert CR and NL combos to single space
    psz = pszMsgStart;
    while (*psz) {
      if (*psz == '\n' || *psz == '\r') {
	while (*(psz + 1) == '\n' || *(psz + 1) == '\r')
	  memmove(psz, psz + 1, strlen(psz));
	*psz = ' ';
      }
      else
	psz++;
    }
  }

  WriteMsg(szMsg);

} // Dos_Error

/**
 * Report runtime logic error using passed message string
 */

VOID Runtime_Error(PCSZ pszSrcFile, UINT cSrcLineNo, PCSZ pszFmt, ...)
{
  CHAR szMsg[4096];
  va_list va;

  // Format caller's message
  va_start(va, pszFmt);
  szMsg[sizeof(szMsg) - 1] = 0;
  vsprintf(szMsg, pszFmt, va);
  va_end(va);

  if (szMsg[sizeof(szMsg) - 1]) {
    fprintf(stdout, "Buffer overflow in Runtime_Error - need %u bytes.\n",
	    strlen(szMsg) + 1);
    fflush(stdout);
  }

  if (strchr(szMsg, ' ') == NULL)
    strcat(szMsg, " failed.");		// Assume simple function name

  WriteMsg(szMsg);

} // Runtime_Error

/*
 * Shut down
 * @param sStatus is exit status
 */

VOID Shut(SHORT sStatus)
{
  exit(sStatus);

} // Shut

/**
 * Format message and output to stdout and log file supplying trailing NL
 * FIXME log file output #if disabled
 * @param pszFmt points for printf format string or NULL
 * @note Local errors reported directly to stdout
 * @note Writes blank line if pszFmt NULL
 */

VOID WriteMsg(PCSZ pszFmt, ...)
{
  va_list va;
  CHAR szMsg[4096];
  UINT c;
  PSZ psz;
  static volatile ULONG cBusy;		// Need to be MT-safe
  static time_t logTime;
  static CHAR szLogTime[CCHMAXPATH];

  if (!pszFmt || !*pszFmt)
    szMsg[0] = 0;			// Want blank line
  else {
    szMsg[sizeof(szMsg) - 1] = 0;	// Overflow check

    // Format caller's message
    va_start(va, pszFmt);
    vsprintf(szMsg, pszFmt, va);
    va_end(va);

    if (szMsg[sizeof(szMsg) - 1]) {
      WriteMsg("WriteMsg buffer overflow (%u/%u)",
	       strlen(szMsg), sizeof(szMsg));
    }
  }

  // Serialize output to be SMP safe
  for (c = 0; ; c++) {
    if (SMPSafeInc(&cBusy), cBusy == 1)
      break;
    SMPSafeDec(&cBusy);
    // Hold off 1 cycle before reporting since some contention expected
    if (c == 1 && Opts.cDebug)
      fputs("WriteMsg waiting for other thread", stdout);
    DosSleep(1);			// Let current owner finish
  }

  fputs(szMsg, stdout);
  fputc('\n', stdout);
  fflush(stdout);

# if 0 // FIXME to be gone
  if (!hLogFile) {
    if (!*szLogFileName) {
      psz = getenv("LOGFILES");
      if (!psz)
	psz = getenv("TMP");
      if (!psz)
	psz = szLogFileName;
      else {
	strcpy(szLogFileName, psz);
	psz = szLogFileName;
	c = strlen(psz);
	if (c > 0 && psz[c - 1] != '\\') {
	  psz += c;
	  *psz++ = '\\';
	}
	else
	  psz += c;
      }
      strcpy(psz, LOGFILE_NAME);
    }
    hLogFile = fopen(szLogFileName, "a");
    if (!hLogFile) {
      static UINT warned;
      c = errno;
      if (!warned) {
	printf("Can not open %s for append (%u)\n", szLogFileName, errno);
	warned = TRUE;
      }
    }
  }
  if (hLogFile) {
    // Formatted timestamp
    time_t now = time(NULL);
    if (now != logTime) {
      struct tm *tm = localtime(&now);
      logTime = now;
      // yyyy/mm/dd hh:mm:ss
      sprintf(szLogTime,
	      "%u/%02u/%02u %02u:%02u:%02u ",
	       tm->tm_year + 1900,
	       tm->tm_mon + 1,		// 0..11
	       tm->tm_mday,
	       tm->tm_hour,
	       tm->tm_min,
	       tm->tm_sec);
    }
    if (*szMsg) {
      fputs(szLogTime, hLogFile);
      // Formatted message followed by NL
      fputs(szMsg, hLogFile);
    }
    fputc('\n', hLogFile);
  }
# endif

  cBusy--;

} // WriteMsg

// eof
